@@ -3,10 +3,33 @@ Use of this source code is governed by the MIT license found in the LICENSE file
 
 General node object to control associated plugwise nodes like: Circle+, Circle, Scan, Stealth
 """
-from plugwise.constants import *
+from datetime import datetime
+from plugwise.constants import (
+    CALLBACK_ALL,
+    CALLBACK_POWER,
+    CALLBACK_RELAY,
+    HA_SWITCH,
+    MAX_TIME_DRIFT,
+    NODE_TYPE_CIRCLE,
+    NODE_TYPE_CIRCLE_PLUS,
+    NODE_TYPE_SCAN,
+    NODE_TYPE_SENSE,
+    NODE_TYPE_STEALTH,
+    NODE_TYPE_SWITCH,
+    NODE_TYPE_STICK,
+)
 from plugwise.message import PlugwiseMessage
-from plugwise.messages.responses import NodeInfoResponse, NodePingResponse
-from plugwise.messages.requests import NodeInfoRequest, NodePingRequest
+from plugwise.messages.responses import (
+    NodeClockResponse,
+    NodeInfoResponse,
+    NodePingResponse,
+)
+from plugwise.messages.requests import (
+    NodeClockGetRequest,
+    NodeClockSetRequest,
+    NodeInfoRequest,
+    NodePingRequest,
+)
 from plugwise.util import validate_mac
 
 
@@ -34,7 +57,12 @@ class PlugwiseNode(object):
         self._node_type = None
         self._hardware_version = None
         self._firmware_version = None
-        self._relay_state = None
+        self._relay_state = False
+        self._last_log_address = None
+        self._last_log_collected = False
+        self._last_info_message = None
+        self._clock_offset = None
+        self.get_clock(self.sync_clock)
 
     def get_available(self) -> bool:
         return self._available
@@ -44,24 +72,18 @@ class PlugwiseNode(object):
             if self._available == False:
                 self._available = True
                 self.stick.logger.debug(
-                    "Mark node %s available",
-                    self.mac.decode("ascii"),
+                    "Mark node %s available", self.get_mac(),
                 )
-                if CALLBACK_ALL in self._callbacks:
-                    for callback in self._callbacks[CALLBACK_ALL]:
-                        callback(None)
+                self._do_all_callbacks()
                 if request_info:
                     self._request_info()
         else:
             if self._available == True:
                 self._available = False
                 self.stick.logger.debug(
-                    "Mark node %s unavailable",
-                    self.mac.decode("ascii"),
+                    "Mark node %s unavailable", self.get_mac(),
                 )
-                if CALLBACK_ALL in self._callbacks:
-                    for callback in self._callbacks[CALLBACK_ALL]:
-                        callback(None)
+                self._do_all_callbacks()
 
     def get_mac(self) -> str:
         """Return mac address"""
@@ -104,11 +126,9 @@ class PlugwiseNode(object):
             return str(self._firmware_version)
         return "Unknown"
 
-    def get_last_update(self) -> str:
+    def get_last_update(self) -> datetime:
         """Return  version"""
-        if self.last_update != None:
-            return str(self.last_update)
-        return "Unknown"
+        return self.last_update
 
     def get_in_RSSI(self) -> int:
         """Return inbound RSSI level"""
@@ -150,20 +170,23 @@ class PlugwiseNode(object):
                 self.stick.logger.debug(
                     "Last update %s of node %s, last message %s",
                     str(self.last_update),
-                    self.mac.decode("ascii"),
+                    self.get_mac(),
                     str(message.timestamp),
                 )
                 self.last_update = message.timestamp
-            if isinstance(message, NodeInfoResponse):
-                self.set_available(True)
-                self._process_info_response(message)
-                self.stick.message_processed(message.seq_id)
-            elif isinstance(message, NodePingResponse):
+            if isinstance(message, NodePingResponse):
                 self.in_RSSI = message.in_RSSI.value
                 self.out_RSSI = message.out_RSSI.value
                 self.ping_ms = message.ping_ms.value
-                self.stick.message_processed(message.seq_id)
                 self.set_available(True, True)
+                self.stick.message_processed(message.seq_id)
+            elif isinstance(message, NodeInfoResponse):
+                self.set_available(True)
+                self._process_info_response(message)
+                self.stick.message_processed(message.seq_id)
+            elif isinstance(message, NodeClockResponse):
+                self._response_clock(message)
+                self.stick.message_processed(message.seq_id)
             else:
                 self.set_available(True)
                 self._on_message(message)
@@ -171,25 +194,109 @@ class PlugwiseNode(object):
             self.stick.logger.debug(
                 "Skip message, mac of node (%s) != mac at message (%s)",
                 message.mac.decode("ascii"),
-                self.mac.decode("ascii"),
+                self.get_mac(),
             )
 
     def _on_message(self, message):
         pass
 
+    def register_callback(self, callback, state=CALLBACK_ALL):
+        """ Register callback to execute when state change happens """
+        if state == CALLBACK_RELAY:
+            if CALLBACK_RELAY not in self._callbacks:
+                self._callbacks[CALLBACK_RELAY] = []
+            self._callbacks[CALLBACK_RELAY].append(callback)
+        if state == CALLBACK_POWER:
+            if CALLBACK_POWER not in self._callbacks:
+                self._callbacks[CALLBACK_POWER] = []
+            self._callbacks[CALLBACK_POWER].append(callback)
+        if state == CALLBACK_ALL:
+            if CALLBACK_ALL not in self._callbacks:
+                self._callbacks[CALLBACK_ALL] = []
+            self._callbacks[CALLBACK_ALL].append(callback)
+
+    def _do_all_callbacks(self):
+        """ Execute callbacks registered for all updates """
+        if CALLBACK_ALL in self._callbacks:
+            for callback in self._callbacks[CALLBACK_ALL]:
+                try:
+                    callback(None)
+                except Exception as e:
+                    self.stick.logger.error(
+                        "Error while executing all callback : %s",
+                        e,
+                    )
+
     def _process_info_response(self, message):
         """ Process info response message"""
         self.stick.logger.debug(
-            "Response info message for plug with mac " + self.mac.decode("ascii")
+            "Response info message for plug %s", self.get_mac()
         )
         if message.relay_state.serialize() == b"01":
-            self._relay_state = True
+            if not self._relay_state:
+                self._relay_state = True
+                self._do_all_callbacks()
         else:
-            self._relay_state = False
+            if self._relay_state:
+                self._relay_state = False
+                self._do_all_callbacks()
         self._hardware_version = int(message.hw_ver.value)
         self._firmware_version = message.fw_ver.value
         self._node_type = message.node_type.value
-        self.stick.logger.debug("Node type        = " + self.get_node_type())
-        self.stick.logger.debug("Relay state      = " + str(self._relay_state))
-        self.stick.logger.debug("Hardware version = " + str(self._hardware_version))
-        self.stick.logger.debug("Firmware version = " + str(self._firmware_version))
+        self._last_info_message = message.timestamp
+        if self._last_log_address != message.last_logaddr.value:
+            self._last_log_address = message.last_logaddr.value
+            self._last_log_collected = False
+        self.stick.logger.debug("Node type        = %s", self.get_node_type())
+        self.stick.logger.debug("Relay state      = %s", str(self._relay_state))
+        self.stick.logger.debug("Hardware version = %s", str(self._hardware_version))
+        self.stick.logger.debug("Firmware version = %s", str(self._firmware_version))
+
+    def _request_power_buffer(self, log_address=None, callback=None):
+        pass
+
+    def get_clock(self, callback=None):
+        """ get current datetime of internal clock of CirclePlus """
+        self.stick.send(
+            NodeClockGetRequest(self.mac), callback,
+        )
+
+    def _response_clock(self, message):
+        dt = datetime(
+            datetime.now().year,
+            datetime.now().month,
+            datetime.now().day,
+            message.time.value.hour,
+            message.time.value.minute,
+            message.time.value.second,
+        )
+        clock_offset = message.timestamp.replace(microsecond=0) - (dt + self.stick.timezone_delta)
+        if clock_offset.days == -1:
+            self._clock_offset = clock_offset.seconds - 86400
+        else:
+            self._clock_offset = clock_offset.seconds
+        self.stick.logger.debug(
+            "Clock of node %s has drifted %s sec",
+            self.get_mac(),
+            str(self._clock_offset),
+        )
+
+    def set_clock(self, callback=None):
+        """ set internal clock of CirclePlus """
+        self.stick.send(
+            NodeClockSetRequest(self.mac, datetime.utcnow()), callback,
+        )
+
+    def sync_clock(self, max_drift=0):
+        """ Resync clock of node if time has drifted more than MAX_TIME_DRIFT
+        """
+        if self._clock_offset != None:
+            if max_drift == 0:
+                max_drift = MAX_TIME_DRIFT
+            if (self._clock_offset > max_drift) or (self._clock_offset < -(max_drift)):
+                self.stick.logger.warning(
+                    "Reset clock of node %s because time has drifted %s sec",
+                    self.get_mac(),
+                    str(self._clock_offset),
+                )
+                self.set_clock()
